"""Query state and runner."""

import asyncio
import hashlib
import json
import logging
import math
import os
import re
import sys
import tempfile
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from aio_overpass.error import (
    ClientError,
    GiveupError,
    QueryLanguageError,
    QueryRejectCause,
    QueryRejectError,
    ResponseError,
    ServerError,
)


__docformat__ = "google"
__all__ = (
    "Query",
    "QueryRunner",
    "DefaultQueryRunner",
    "RequestTimeout",
    "DEFAULT_MAXSIZE",
    "DEFAULT_TIMEOUT",
)


DEFAULT_MAXSIZE = 512
"""Default ``maxsize`` setting in mebibytes."""

DEFAULT_TIMEOUT = 180
"""Default ``timeout`` setting in seconds."""

_COPYRIGHT = "The data included in this document is from www.openstreetmap.org. The data is made available under ODbL."  # noqa: E501
"""This is the same copyright notice included in result sets"""

_SETTING_PATTERN = re.compile(r"\[(\w+?):(.+?)]\s*;?")
"""A pattern to match setting declarations (not the entire settings statement)."""

_NULL_LOGGER = logging.getLogger()
_NULL_LOGGER.addHandler(logging.NullHandler())


class Query:
    """
    State of a query that is either pending, running, successful, or failed.

    Args:
        input_code: The input Overpass QL code. Note that some settings might be changed
                    by query runners, notably the 'timeout' and 'maxsize' settings.
        logger: The logger to use for all logging output related to this query.
        **kwargs: Additional keyword arguments that can be used to identify queries.

    References:
        - https://wiki.openstreetmap.org/wiki/Overpass_API/Overpass_QL
    """

    __slots__ = (
        "_error",
        "_input_code",
        "_kwargs",
        "_logger",
        "_nb_tries",
        "_request_timeout",
        "_response",
        "_response_bytes",
        "_run_timeout_secs",
        "_settings",
        "_time_end_try",
        "_time_start",
        "_time_start_req",
        "_time_start_try",
    )

    def __init__(self, input_code: str, logger: logging.Logger = _NULL_LOGGER, **kwargs) -> None:
        self._input_code = input_code
        self._logger = logger
        self._kwargs = kwargs

        self._settings = dict(_SETTING_PATTERN.findall(input_code))

        self._settings["out"] = "json"

        if "maxsize" not in self._settings:
            self._settings["maxsize"] = DEFAULT_MAXSIZE * 1024 * 1024

        if "timeout" not in self._settings:
            self._settings["timeout"] = DEFAULT_TIMEOUT

        self._run_timeout_secs: float | None = None
        self._request_timeout: RequestTimeout = RequestTimeout()

        # set by the client that executes this query
        self._error: ClientError | None = None
        self._response: dict | None = None
        self._response_bytes = 0.0
        self._nb_tries = 0

        self._time_start: _Instant | None = None
        """time prior to executing the first try"""

        self._time_start_try: _Instant | None = None
        """time prior to executing the most recent try"""

        self._time_start_req: _Instant | None = None
        """time prior to executing the most recent try's query request"""

        self._time_end_try: _Instant | None = None
        """time the most recent try finished"""

    def reset(self) -> None:
        """Reset the query to its initial state, ignoring previous tries."""
        Query.__init__(self, input_code=self._input_code, **self._kwargs)

    @property
    def input_code(self) -> str:
        """The original input Overpass QL source code."""
        return self._input_code

    @property
    def kwargs(self) -> dict:
        """
        Keyword arguments that can be used to identify queries.

        The default query runner will log these values when a query is run.
        """
        return self._kwargs

    @property
    def logger(self) -> logging.Logger:
        """The logger used for logging output related to this query."""
        return self._logger

    @property
    def nb_tries(self) -> int:
        """Current number of tries."""
        return self._nb_tries

    @property
    def error(self) -> ClientError | None:
        """
        Error of the most recent try.

        Returns:
            an error or ``None`` if the query wasn't tried or hasn't failed
        """
        return self._error

    @property
    def response(self) -> dict | None:
        """
        The entire JSON response of the query.

        Returns:
            the response, or ``None`` if the query has not successfully finished (yet)
        """
        return self._response

    @property
    def was_cached(self) -> bool | None:
        """
        Indicates whether the query result was cached.

        Returns:
            ``None`` if the query has not been run yet.
            ``True`` if the query has a result set with zero tries.
            ``False`` otherwise.
        """
        if self._response is None:
            return None
        return self._nb_tries == 0

    @property
    def result_set(self) -> list[dict] | None:
        """
        The result set of the query.

        This is open data, licensed under the Open Data Commons Open Database License (ODbL).
        You are free to copy, distribute, transmit and adapt this data, as long as you credit
        OpenStreetMap and its contributors. If you alter or build upon this data, you may
        distribute the result only under the same licence.

        Returns:
            the elements of the result set, or ``None`` if the query has not successfully
            finished (yet)

        References:
            - https://www.openstreetmap.org/copyright
            - https://opendatacommons.org/licenses/odbl/1-0/
        """
        if not self._response:
            return None
        return self._response["elements"]

    @property
    def response_size_mib(self) -> float | None:
        """
        The size of the response in mebibytes.

        Returns:
            the size, or ``None`` if the query has not successfully finished (yet)
        """
        if self._response is None:
            return None
        return self._response_bytes / 1024.0 / 1024.0

    @property
    def maxsize_mib(self) -> float:
        """
        The current value of the [maxsize:*] setting in mebibytes.

        This size indicates the maximum allowed memory for the query in bytes RAM on the server,
        as expected by the user. If the query needs more RAM than this value, the server may abort
        the query with a memory exhaustion. The higher this size, the more probably the server
        rejects the query before executing it.
        """
        return float(self._settings["maxsize"]) // 1024.0 // 1024.0

    @maxsize_mib.setter
    def maxsize_mib(self, value: float) -> None:
        if value <= 0.0:
            msg = "maxsize_mib must be > 0.0"
            raise ValueError(msg)
        self._settings["maxsize"] = int(value * 1024.0 * 1024.0)

    @property
    def timeout_secs(self) -> int:
        """
        The current value of the [timeout:*] setting in seconds.

        This duration is the maximum allowed runtime for the query in seconds, as expected by the
        user. If the query runs longer than this time, the server may abort the query. The higher
        this duration, the more probably the server rejects the query before executing it.
        """
        return int(self._settings["timeout"])

    @timeout_secs.setter
    def timeout_secs(self, value: int) -> None:
        if value < 1:
            msg = "timeout_secs must be >= 1"
            raise ValueError(msg)
        self._settings["timeout"] = value

    @property
    def run_timeout_secs(self) -> float | None:
        """
        A limit to ``run_duration_secs``, that cancels the query when exceeded.

        Defaults to no timeout.

        The client will raise a ``QueryCancelledError`` if the timeout is reached.

        Not to be confused with ``timeout_secs``, which is a setting for the Overpass API instance,
        that limits the query execution time. Instead, this value can be used to limit the total
        client-side time spent on this query (see ``Client.run_query``).
        """
        return self._run_timeout_secs

    @run_timeout_secs.setter
    def run_timeout_secs(self, value: float | None) -> None:
        if value is not None and value <= 0.0:
            msg = "run_timeout_secs must be > 0"
            raise ValueError(msg)
        self._run_timeout_secs = value

    @property
    def run_timeout_elapsed(self) -> bool:
        """Returns ``True`` if ``run_timeout_secs`` is set and has elapsed."""
        return (
            self.run_timeout_secs is not None
            and self.run_duration_secs is not None
            and self.run_timeout_secs < self.run_duration_secs
        )

    @property
    def request_timeout(self) -> "RequestTimeout":
        """Request timeout settings for this query."""
        return self._request_timeout

    @request_timeout.setter
    def request_timeout(self, value: "RequestTimeout") -> None:
        self._request_timeout = value

    @property
    def code(self) -> str:
        """
        The Overpass QL source code of this query.

        This is different from ``input_code`` only when it comes to settings.
        """
        return self._code(without_dynamic_settings=False)

    def _code(self, without_dynamic_settings: bool) -> str:
        settings_iter = iter(self._settings.items())

        if without_dynamic_settings:
            settings_iter = ((k, v) for k, v in settings_iter if k not in {"maxsize", "timeout"})

        settings = "".join((f"[{k}:{v}]" for k, v in settings_iter)) + ";"

        # Remove the original settings statement
        code = _SETTING_PATTERN.sub("", self._input_code)

        return f"{settings}\n{code}"

    @property
    def cache_key(self) -> str:
        """
        Hash QL code, and return its digest as hexadecimal string.

        The default query runner uses this as cache key.
        """
        # hash QL code without [timeout:*] and [maxsize:*] settings
        code = self._code(without_dynamic_settings=True)
        return hashlib.sha256(code.encode("utf-8")).hexdigest()

    @property
    def done(self) -> bool:
        """Returns ``True`` if the result set was received."""
        return self._response is not None

    @property
    def request_duration_secs(self) -> float | None:
        """
        How long it took to fetch the result set in seconds.

        This is the duration starting with the API request, and ending once
        the result is written to this query object. Although it depends on how busy
        the API instance is, this can give some indication of how long a query takes.

        Returns:
            the duration or ``None`` if there is no result set yet, or when it was cached.
        """
        if self._response is None or self.was_cached:
            return None

        assert self._time_end_try is not None
        assert self._time_start_req is not None

        return self._time_end_try - self._time_start_req

    @property
    def run_duration_secs(self) -> float | None:
        """
        The total required time for this query in seconds (so far).

        Returns:
            the duration or ``None`` if there is no result set yet, or when it was cached.
        """
        if self._time_start is None:
            return None

        if self._time_end_try:
            return self._time_end_try - self._time_start

        return self._time_start.elapsed_secs_since

    @property
    def api_version(self) -> str | None:
        """
        The Overpass API version used by the queried instance.

        Returns:
            f.e. ``"Overpass API 0.7.56.8 7d656e78"``, or ``None`` if the query
            has not successfully finished (yet)

        References:
            - https://wiki.openstreetmap.org/wiki/Overpass_API/versions
        """
        if self._response is None:
            return None

        return self._response["generator"]

    @property
    def timestamp_osm(self) -> datetime | None:
        """
        All OSM edits that have been uploaded before this date are included.

        It can take a couple of minutes for changes to the database to show up in the
        Overpass API query results.

        Returns:
            the timestamp, or ``None`` if the query has not successfully finished (yet)
        """
        if self._response is None:
            return None

        date_str = self._response["osm3s"]["timestamp_osm_base"]
        return datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%SZ")

    @property
    def timestamp_areas(self) -> datetime | None:
        """
        All area data edits that have been uploaded before this date are included.

        If the query involves area data processing, this is the date of the latest edit
        that has been considered in the most recent batch run of the area generation.

        Returns:
            the timestamp, or ``None`` if the query has not successfully finished (yet), or
            if it does not involve area data processing.
        """
        if self._response is None:
            return None

        date_str = self._response["osm3s"].get("timestamp_areas_base")
        if not date_str:
            return None

        return datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%SZ")

    @property
    def copyright(self) -> str:
        """A copyright notice that comes with the result set."""
        if self._response is None:
            return _COPYRIGHT

        return self._response["osm3s"].get("copyright") or _COPYRIGHT

    def __str__(self) -> str:
        query = f"query {self.kwargs}" if self.kwargs else "query <no kwargs>"

        size = self.response_size_mib
        time_request = self.request_duration_secs
        time_total = self.run_duration_secs

        if self.nb_tries == 0:
            details = "pending"
        elif self.done:
            if self.nb_tries == 1:
                details = f"{size}mb in {time_request:.01f}s"
            else:
                details = f"{size}mb in {time_request:.01f} ({time_total:.01f})s"
        else:
            t = "try" if self.nb_tries == 1 else "tries"
            details = f"failing after {self.nb_tries} {t}, {time_total:.01f}s"

        return f"{query} ({details})"

    def __repr__(self) -> str:
        cls_name = type(self).__name__

        details = {
            "kwargs": self._kwargs,
            "done": self.done,
        }

        if self.nb_tries == 0 or self.error:
            details["tries"] = self.nb_tries

        if self.error:
            details["error"] = type(self.error).__name__

        if self.done:
            details["response_size"] = f"{self.response_size_mib:.02f}mb"

            if not self.was_cached:
                details["request_duration"] = f"{self.request_duration_secs:.02f}s"

        if self.nb_tries > 0:
            details["run_duration"] = f"{self.run_duration_secs:.02f}s"

        details_str = ", ".join((f"{k}={v!r}" for k, v in details.items()))

        return f"{cls_name}({details_str})"

    def _mutator(self) -> "_QueryMutator":
        return _QueryMutator(self)


class _QueryMutator:
    __slots__ = ("_query",)

    def __init__(self, query: Query) -> None:
        self._query = query

    def begin_try(self) -> None:
        if self._query._time_start is None:
            self._query._time_start = _Instant.now()

        self._query._time_start_try = _Instant.now()
        self._query._time_start_req = None
        self._query._time_end_try = None

    def begin_request(self) -> None:
        self._query._time_start_req = _Instant.now()

    def succeed_try(self, response: dict, response_bytes: int) -> None:
        self._query._time_end_try = _Instant.now()
        self._query._response = response
        self._query._response_bytes = response_bytes
        self._query._error = None

    def fail_try(self, err: ClientError) -> None:
        self._query._error = err

    def end_try(self) -> None:
        self._query._nb_tries += 1


@dataclass(slots=True, frozen=True, repr=False, order=True)
class _Instant:
    """
    Measurement of a monotonic clock.

    Attributes:
        when: the current time, according to the event loop's internal monotonic clock
              (details are unspecified and may differ per event loop).
    """

    when: float

    @classmethod
    def now(cls) -> "_Instant":
        return cls(when=asyncio.get_event_loop().time())

    @property
    def ceil(self) -> int:
        return math.ceil(self.when)

    @property
    def elapsed_secs_since(self) -> float:
        return asyncio.get_event_loop().time() - self.when

    def __sub__(self, earlier: "_Instant") -> float:
        if self.when < earlier.when:
            msg = f"{self} is earlier than {earlier}"
            raise ValueError(msg)
        return self.when - earlier.when

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.when:.02f})"


@dataclass(slots=True)
class RequestTimeout:
    """
    Request timeout settings.

    Attributes:
        total_without_query_secs: If set, the sum of this duration and the query's ``[timeout:*]``
                                  setting is used as timeout duration of the entire request,
                                  including connection establishment, request sending and response
                                  reading (``aiohttp.ClientTimeout.total``).
                                  Defaults to 20 seconds.
        sock_connect_secs: The maximum number of seconds allowed for pure socket connection
                           establishment (same as ``aiohttp.ClientTimeout.sock_connect``).
        each_sock_read_secs: The maximum number of seconds allowed for the period between reading
                             a new chunk of data (same as ``aiohttp.ClientTimeout.sock_read``).
    """

    total_without_query_secs: float | None = 20.0
    sock_connect_secs: float | None = None
    each_sock_read_secs: float | None = None

    def __post_init__(self) -> None:
        if self.total_without_query_secs is not None and self.total_without_query_secs <= 0.0:
            msg = "'total_without_query_secs' has to be > 0"
            raise ValueError(msg)

        if self.sock_connect_secs is not None and self.sock_connect_secs <= 0.0:
            msg = "'sock_connect_secs' has to be > 0"
            raise ValueError(msg)

        if self.each_sock_read_secs is not None and self.each_sock_read_secs <= 0.0:
            msg = "'each_sock_read_secs' has to be > 0"
            raise ValueError(msg)


class QueryRunner(ABC):
    """
    A query runner is an async function that is called before a client makes an API request.

    Query runners can be used to…
     - …retry queries when they fail
     - …modify queries, f.e. to lower/increase their maxsize/timeout
     - …log query results & errors
     - …implement caching

    The absolute minimum a query runner function has to do is to simply return to (re)try
    a query, or to raise ``query.error`` to stop trying.
    """

    __slots__ = ()

    @abstractmethod
    async def __call__(self, query: Query) -> None:
        """Called with the current query state before the client makes an API request."""
        pass


class DefaultQueryRunner(QueryRunner):
    """
    The default query runner.

    This runner…
     - …retries with an increasing back-off period in between tries if the server is too busy
     - …retries and doubles timeout and maxsize settings if they were exceeded
     - …limits the number of tries
     - …optionally caches query results in temp files

    This runner does *not* lower timeout and maxsize settings if the server rejected a query.

    Args:
        max_tries: The maximum number of times a query is tried. (5 by default)
        cache_ttl_secs: Amount of seconds a query's result set is cached for.
                        Set to zero to disable caching. (zero by default)
    """

    __slots__ = (
        "_max_tries",
        "_cache_ttl_secs",
    )

    def __init__(self, max_tries: int = 5, cache_ttl_secs: int = 0) -> None:
        if max_tries < 1:
            msg = "max_tries must be >= 1"
            raise ValueError(msg)

        if cache_ttl_secs < 0:
            msg = "cache_ttl_secs must be >= 0"
            raise ValueError(msg)

        self._max_tries = max_tries
        self._cache_ttl_secs = cache_ttl_secs

    def _cache_read(self, query: Query) -> None:
        logger = query.logger

        if _FORCE_DISABLE_CACHE:
            logger.debug("caching is forced disabled")
            return
        if not self._cache_ttl_secs:
            logger.debug("caching is disabled")
            return

        now = int(time.time())

        file_name = f"{query.cache_key}.json"
        file_path = Path(tempfile.gettempdir()) / file_name

        if not file_path.exists():
            logger.info("result was not cached")
            logger.debug(f"checked for cache at {file_path}")
            return

        try:
            with open(file_path, mode="r", encoding="utf-8") as file:
                response = json.load(file)
        except (OSError, json.JSONDecodeError):
            logger.exception(f"failed to read cached {query}")
            return

        if response.get(_EXPIRATION_KEY, 0) <= now:
            logger.info(f"{query} cache expired")
            return

        query._response = response
        logger.info(f"{query} was cached")

    def _cache_write(self, query: Query) -> None:
        logger = query.logger

        if not self._cache_ttl_secs or _FORCE_DISABLE_CACHE:
            return

        now = int(time.time())

        assert query._response is not None
        query._response[_EXPIRATION_KEY] = now + self._cache_ttl_secs

        file_name = f"{query.cache_key}.json"
        file_path = Path(tempfile.gettempdir()) / file_name

        logger.debug(f"caching at {file_path}…")

        try:
            with open(file_path, mode="w", encoding="utf-8") as file:
                json.dump(query._response, file)
        except OSError:
            logger.exception(f"failed to cache {query}")

    async def __call__(self, query: Query) -> None:
        """Called with the current query state before the client makes an API request."""
        logger = query.logger

        # Check cache ahead of first try
        if query.nb_tries == 0:
            await asyncio.to_thread(self._cache_read, query)

        # Success or cached
        if query.done:
            if not query.was_cached:
                await asyncio.to_thread(self._cache_write, query)
            return

        err = query.error

        # Do not retry if we exhausted all tries, when a retry would not change the result,
        # or when the timeout was reached.
        failed = query.nb_tries == self._max_tries or isinstance(
            err, ResponseError | QueryLanguageError | GiveupError
        )

        if isinstance(err, ServerError):
            logger.error(f"server error body:\n{err.body}")

        # Exhausted all tries; do not retry.
        if err and failed:
            logger.error(f"give up on {query}", exc_info=err)

            if isinstance(err, ResponseError):
                logger.error(f"unexpected response body:\n{err.body}")

            raise err

        if isinstance(err, QueryRejectError):
            # Wait if the server is too busy.
            if err.cause == QueryRejectCause.TOO_BUSY:
                backoff = _fibo_backoff_secs(query.nb_tries)
                logger.info(f"retry {query} in {backoff:.1f}s")
                await asyncio.sleep(backoff)

            # Wait until a slot opens if the rate limit was exceeded.
            elif err.cause == QueryRejectCause.TOO_MANY_QUERIES:
                pass  # let client enforce cooldown

            # Double timeout if exceeded.
            elif err.cause == QueryRejectCause.EXCEEDED_TIMEOUT:
                old = f"{query.timeout_secs:.1f}s"
                query.timeout_secs *= 2
                new = f"{query.timeout_secs:.1f}s"
                logger.info(f"increased [timeout:*] for {query} from {old} to {new}")

            # Double maxsize if exceeded.
            elif err.cause == QueryRejectCause.EXCEEDED_MAXSIZE:
                old = f"{query.maxsize_mib:.1f}mib"
                query.maxsize_mib *= 2
                new = f"{query.maxsize_mib:.1f}mib"
                logger.info(f"increased [maxsize:*] for {query} from {old} to {new}")


def _fibo_backoff_secs(tries: int) -> float:
    """Fibonacci sequence without zero: 1, 1, 2, 3, 5, 8, etc."""
    a, b = 1.0, 1.0

    for _ in range(tries):
        a, b = b, a + b

    return a


def __cache_delete(query: Query) -> None:
    """Clear a response cached by the default runner (only to be used in tests)."""
    file_name = f"{query.cache_key}.json"
    file_path = Path(tempfile.gettempdir()) / file_name
    file_path.unlink(missing_ok=True)


def __cache_expire(query: Query) -> None:
    """Clear a response cached by the default runner (only to be used in tests)."""
    file_name = f"{query.cache_key}.json"
    file_path = Path(tempfile.gettempdir()) / file_name

    with open(file_path, mode="r", encoding="utf-8") as file:
        response = json.load(file)

    response[_EXPIRATION_KEY] = 0

    with open(file_path, mode="w", encoding="utf-8") as file:
        json.dump(response, file)


_EXPIRATION_KEY = "__expiration__"
_IS_CI = os.getenv("GITHUB_ACTIONS") == "true"
_IS_UNIT_TEST = "pytest" in sys.modules
_FORCE_DISABLE_CACHE = _IS_CI and not _IS_UNIT_TEST



"""
Error types.

TODO add ServerError
```
                            (ClientError)
                                  ╷
     ┌──────────────┬─────────────┼────────────┬────────────────┬────────────────┐
     ╵              ╵             ╵            ╵                ╵                ╵
ServerError    RunnerError   (QueryError)   GiveupError    ResponseError     CallError
                                  ╷                             ╷                ╷
         ┌────────────────────────┼──────────────────────┐      │                │
         ╵                        ╵                      ╵      ╵                ╵
QueryLanguageError         QueryRejectError          QueryResponseError   CallTimeoutError
```
"""
import asyncio
import html
import re
from dataclasses import dataclass
from enum import Enum, auto
from json import JSONDecodeError
from typing import no_type_check

import aiohttp
import aiohttp.typedefs


__docformat__ = "google"
__all__ = (
    "ClientError",
    "CallError",
    "CallTimeoutError",
    "ServerError",
    "ResponseError",
    "GiveupError",
    "QueryError",
    "QueryLanguageError",
    "QueryRejectError",
    "QueryRejectCause",
    "QueryResponseError",
    "RunnerError",
)


class ClientError(Exception):
    """Base exception for failed Overpass API requests and queries."""


@dataclass
class RunnerError(ClientError):
    """
    The query runner raised an exception.

    This is an unexpected error, in contrast to the runner intentionally
    raising ``query.error`` to stop retrying. The cause of this error
    is therefore anything but a ``ClientError``.

    Attributes:
        cause: the exception that caused this error
    """

    cause: BaseException

    def __post_init__(self) -> None:
        # imitate "raise RunnerError(...) from cause"
        self.__cause__ = self.cause

    def __str__(self) -> str:
        return str(self.cause)


@dataclass
class CallError(ClientError):
    """
    Failed to make an API request.

    This error is raised when the client failed to get any response,
    f.e. due to connection issues.

    Attributes:
        cause: the exception that caused this error
    """

    cause: aiohttp.ClientError

    def __post_init__(self) -> None:
        # imitate "raise CallError(...) from cause"
        self.__cause__ = self.cause

    def __str__(self) -> str:
        return str(self.cause)


@dataclass
class CallTimeoutError(CallError):
    """
    An API request timed out.

    Attributes:
        cause: the exception that caused this error
        after_secs: the configured timeout for the request
    """

    cause: asyncio.TimeoutError  # type: ignore[assignment]
    after_secs: float

    def __post_init__(self) -> None:
        # imitate "raise CallTimeoutError(...) from cause"
        self.__cause__ = self.cause

    def __str__(self) -> str:
        return str(self.cause)


@dataclass
class ResponseError(ClientError):
    """
    Unexpected API response.

    This error is raised when a request fails, but we can't specifically say why.
    This may also indicate a bug on our end, since the client is meant
    to process almost any response of an Overpass API instance.

    Attributes:
        response: the unexpected response
        body: the response body
        cause: an optional exception that may have caused this error
    """

    response: aiohttp.ClientResponse
    body: str
    cause: aiohttp.ClientResponseError | JSONDecodeError | None

    def __post_init__(self) -> None:
        # imitate "raise ResponseError(...) from cause"
        self.__cause__ = self.cause

    def __str__(self) -> str:
        if self.cause is None:
            return f"unexpected response: {self.response.status}"
        return f"unexpected response: {self.response.status}: {self.cause}"


@dataclass
class ServerError(ClientError):
    """
    Server error response.

    Responses with status code ``500`` and above are represented by this error class.
    These are errors that happen on the Overpass API instance. It's usually worth
    retrying, since these errors may just be temporary.

    Attributes:
        response: the response
        body: the response body
    """

    response: aiohttp.ClientResponse
    body: str

    def __str__(self) -> str:
        return f"server error: {self.response.status}"


@dataclass
class GiveupError(ClientError):
    """
    The client spent too long running a query, and gave up.

    This error is raised when the run timeout duration set by a query runner
    is or would be exceeded.

    Attributes:
        kwargs: the query's ``kwargs``
        after_secs: the total time spent on the query
    """

    kwargs: dict
    after_secs: float

    def __str__(self) -> str:
        query = f"query {self.kwargs}" if self.kwargs else "query <no kwargs>"
        return f"gave up on {query} after {self.after_secs:.01f} seconds"


@dataclass
class QueryError(ClientError):
    """
    Base exception for queries that failed at the Overpass API server.

    Attributes:
        kwargs: the query's ``kwargs``
        remarks: the error remarks provided by the API
    """

    kwargs: dict
    remarks: list[str]

    def __str__(self) -> str:
        query = f"query {self.kwargs}" if self.kwargs else "query <no kwargs>"
        first = f"'{self.remarks[0]}'"
        rest = f" (+{len(self.remarks) - 1} more)" if len(self.remarks) > 1 else ""
        return f"{query} failed: {first}{rest}"


@dataclass
class QueryResponseError(ResponseError, QueryError):
    """
    Unexpected query response.

    This error is raised when a query fails (thus extends ``QueryError``),
    but we can't specifically say why (thus also extends ``ResponseError``).
    """

    def __str__(self) -> str:
        query = f"query {self.kwargs}" if self.kwargs else "query <no kwargs>"

        if self.remarks:
            first = f"'{self.remarks[0]}'"
            rest = f" (+{len(self.remarks) - 1} more)" if len(self.remarks) > 1 else ""
            return f"{query} failed: {first}{rest}"

        return f"{query} failed with status {self.response.status}"


@dataclass
class QueryLanguageError(QueryError):
    """
    Indicates the query's QL code is not valid.

    Retrying is pointless when encountering this error.
    """


class QueryRejectCause(Enum):
    """Details why a query was rejected or cancelled by an API server."""

    TOO_BUSY = auto()
    """
    Gateway rejection. The server has already so much load that the request cannot be executed.

    Smaller ``[timeout:*]`` and/or ``[maxsize:*]`` values might make the request acceptable.
    """

    TOO_MANY_QUERIES = auto()
    """
    Gateway rejection. There are no open slots for queries from your IP address.

    Running queries take up a slot, and the number of slots is limited. A client
    will only run as many concurrent requests as there are slots, which should make this
    a rare error, assuming you are not making requests through another client.
    """

    EXCEEDED_TIMEOUT = auto()
    """
    Runtime rejection. The query has been (or surely will be) running longer than its proposed
    ``[timeout:*]``, and has been cancelled by the server.

    A higher ``[timeout:*]`` value might allow the query run to completion, but also makes it
    more likely to be rejected by a server under heavy load, before executing it (see ``TOO_BUSY``).
    """

    EXCEEDED_MAXSIZE = auto()
    """
    Runtime rejection. The memory required to execute the query has (or surely will) exceed
    its proposed ``[maxsize:*]``, and has been cancelled by the server.

    A higher ``[maxsize:*]`` value might allow the query run to completion, but also makes it
    more likely to be rejected by a server under heavy load, before executing it (see ``TOO_BUSY``).
    """

    def __str__(self) -> str:
        match self:
            case QueryRejectCause.TOO_BUSY:
                return "server too busy"
            case QueryRejectCause.TOO_MANY_QUERIES:
                return "too many queries"
            case QueryRejectCause.EXCEEDED_TIMEOUT:
                return "exceeded 'timeout'"
            case QueryRejectCause.EXCEEDED_MAXSIZE:
                return "exceeded 'maxsize'"
            case _:
                raise AssertionError


@dataclass
class QueryRejectError(QueryError):
    """
    A query was rejected or cancelled by the API server.

    Attributes:
        cause: why the query was rejected or cancelled
    """

    cause: QueryRejectCause

    def __str__(self) -> str:
        match self.cause:
            case QueryRejectCause.TOO_BUSY | QueryRejectCause.TOO_MANY_QUERIES:
                rejection = "query rejected"
            case QueryRejectCause.EXCEEDED_TIMEOUT | QueryRejectCause.EXCEEDED_MAXSIZE:
                rejection = "query cancelled"
            case _:
                raise AssertionError(self.cause)

        return f"{rejection}: {self.cause}"


@no_type_check
async def _to_client_error(
    obj: aiohttp.ClientResponse | aiohttp.ClientError,
) -> CallError | ResponseError | ServerError:
    """
    Build a ``ClientError`` from either an ``aiohttp`` client error or an unrecognized response.

    Returns:
        - ``CallError`` if ``obj`` is an ``aiohttp.ClientError``,
          but not an ``aiohttp.ClientResponseError``.
        - ``ResponseError`` if the status code is < 500.
        - ``ServerError`` if the status code is >= 500.
    """
    match obj:
        case aiohttp.ClientResponse():
            response = obj
            cause = None
        case aiohttp.ClientResponseError():
            response = obj.history[-1]
            cause = obj
        case _:
            return CallError(cause=obj)  # we didn't get a response

    return await __server_or_response_error(response, cause)


async def _result_or_raise(response: aiohttp.ClientResponse, query_kwargs: dict) -> dict:
    """
    Try to extract the query result set from a response.

    Raises:
        CallError: When there is any sort of connection error.
        RejectError: When encountering "Too Many Requests" or "Gateway Timeout";
                     when there's a JSON remark indicating query rejection or cancellation;
                     when there's an HTML error message indicating query rejection or cancellation.
        QueryError: When there's any other JSON remark or HTML error message.
        ResponseError: When encountering an unexpected response with status code < 500.
        ServerError: When encountering an unexpected response with status code >= 500.
    """
    await __raise_for_plaintext_result(response)

    await __raise_for_html_result(response, query_kwargs)

    return await __raise_for_json_result(response, query_kwargs)


async def __raise_for_json_result(response: aiohttp.ClientResponse, query_kwargs: dict) -> dict:
    try:
        json = await response.json()
        if json is None:
            raise await __server_or_response_error(response, cause=None)
    except aiohttp.ClientResponseError as err:
        raise await __server_or_response_error(response, cause=err) from err
    except JSONDecodeError as err:
        raise await __server_or_response_error(response, cause=err) from err

    if remark := json.get("remark"):
        if timeout_cause := __match_reject_cause(remark):
            raise QueryRejectError(kwargs=query_kwargs, remarks=[remark], cause=timeout_cause)

        raise QueryResponseError(
            response=response,
            body=await response.text(),
            cause=None,
            kwargs=query_kwargs,
            remarks=[remark],
        )

    expected_fields = ("version", "generator", "osm3s", "elements")
    expected_osm3s_fields = ("timestamp_osm_base", "copyright")
    if any(f not in json for f in expected_fields) or any(
        f not in json["osm3s"] for f in expected_osm3s_fields
    ):
        raise await __server_or_response_error(response, cause=None)

    return json


async def __raise_for_html_result(response: aiohttp.ClientResponse, query_kwargs: dict) -> None:
    """
    Raise a fitting exception based on error remarks in an HTML response.

    Raises:
        RejectError: if one of the error remarks indicate that the query was rejected or cancelled
        QueryError: when encountering other error remarks
        ResponseError: when error remarks cannot be extracted from the response
    """
    if response.content_type != "text/html":
        return

    text = await response.text()

    pattern = re.compile("Error</strong>: (.+?)</p>", re.DOTALL)
    errors = [html.unescape(err.strip()) for err in pattern.findall(text)]

    if not errors:  # unexpected format
        raise await __server_or_response_error(response, cause=None)

    if any(__is_ql_error(msg) for msg in errors):
        raise QueryLanguageError(kwargs=query_kwargs, remarks=errors)

    reject_causes = [cause for err in errors if (cause := __match_reject_cause(err))]

    if reject_causes:
        raise QueryRejectError(kwargs=query_kwargs, remarks=errors, cause=reject_causes[0])

    raise QueryResponseError(
        response=response,
        body=await response.text(),
        cause=None,
        kwargs=query_kwargs,
        remarks=errors,
    )


async def __raise_for_plaintext_result(response: aiohttp.ClientResponse) -> None:
    if response.content_type != "text/plain":
        return
    raise await __server_or_response_error(response, cause=None)


def __match_reject_cause(error_msg: str) -> QueryRejectCause | None:
    """
    Check if the given error message indicates that a query was rejected or cancelled.

    AFAIK, neither the 'remarks' in JSON responses, nor the errors listed in HTML responses
    are neatly listed somewhere, but it seems matching a small subset of remarks is enough
    to identify recoverable errors.

    References:
        - Related: https://github.com/DinoTools/python-overpy/issues/62
        - Examples in the API source: https://github.com/drolbr/Overpass-API/search?q=runtime_error
    """
    if "Please check /api/status for the quota of your IP address" in error_msg:
        return QueryRejectCause.TOO_MANY_QUERIES

    if "The server is probably too busy to handle your request" in error_msg:
        return QueryRejectCause.TOO_BUSY

    if "Query timed out" in error_msg:
        return QueryRejectCause.EXCEEDED_TIMEOUT

    if "out of memory" in error_msg:
        return QueryRejectCause.EXCEEDED_MAXSIZE

    return None


def __is_ql_error(error_msg: str) -> bool:
    """Check if the given error message indicates that a query has bad QL code."""
    return (
        "encoding error:" in error_msg
        or "parse error:" in error_msg
        or "static error:" in error_msg
    )


async def __server_or_response_error(
    response: aiohttp.ClientResponse,
    cause: aiohttp.ClientResponseError | JSONDecodeError | None,
) -> ServerError | ResponseError:
    if response.status >= 500:
        return ServerError(
            response=response,
            body=await response.text(),
        )
    return ResponseError(
        response=response,
        body=await response.text(),
        cause=cause,
    )



"""Interface for making API calls."""

import asyncio
import re
from contextlib import suppress
from dataclasses import dataclass
from urllib.parse import urljoin

from aio_overpass import __version__
from aio_overpass.error import (
    CallError,
    CallTimeoutError,
    ClientError,
    GiveupError,
    QueryRejectCause,
    QueryRejectError,
    RunnerError,
    _result_or_raise,
    _to_client_error,
)
from aio_overpass.query import DefaultQueryRunner, Query, QueryRunner

import aiohttp
from aiohttp import ClientTimeout
from aiohttp.helpers import sentinel


__docformat__ = "google"
__all__ = (
    "Client",
    "Status",
    "DEFAULT_INSTANCE",
    "DEFAULT_USER_AGENT",
)


DEFAULT_INSTANCE = "https://overpass-api.de/api/"
"""Main Overpass API instance."""

DEFAULT_USER_AGENT = f"aio-overpass/{__version__} (https://github.com/timwie/aio-overpass)"
"""User agent that points to the ``aio-overpass`` repo."""


@dataclass(slots=True)
class Status:
    """
    Information about the API server's rate limit.

    Attributes:
        slots: The maximum number of concurrent queries per IP
               (or ``None`` if there is no rate limit).
        free_slots: The number of slots open for this IP
                    (or ``None`` if there is no rate limit).
        cooldown_secs: The number of seconds until a slot opens for this IP
                       (or 0 if there is a free slot).
        concurrency: Maximum concurrent queries configured for this client.
        endpoint: Announced endpoint. For example, there are two distinct servers
                  that both can be reached by the main Overpass API instance.
                  Depending on server load, a query may be sent to either of them.
                  This value is the server name, f.e. ``"gall.openstreetmap.de/"``.
        nb_running_queries: Number of currently running queries for this IP.
    """

    slots: int | None
    free_slots: int | None
    cooldown_secs: int
    concurrency: int
    endpoint: str | None
    nb_running_queries: int

    def __repr__(self) -> str:
        f, s, c = self.free_slots, self.slots, self.cooldown_secs

        if self.slots:
            return f"{type(self).__name__}(slots={f}/{s}, cooldown={c}s)"

        return f"{type(self).__name__}(slots=∞, cooldown={c}s)"


class Client:
    """
    A client for the Overpass API.

    Requests are rate-limited according to the configured number of slots per IP for the specified
    API server. By default, queries are retried whenever the server is too busy, or the rate limit
    was exceeded. Custom query runners can be used to implement your own retry strategy.

    Args:
        url: The url of an Overpass API instance. Defaults to the main Overpass API instance.
        user_agent: A string used for the User-Agent header. It is good practice to provide a string
                    that identifies your application, and includes a way to contact you (f.e. an
                    e-mail, or a link to a repository). This is important if you make too many
                    requests, or queries that require a lot of resources.
        concurrency: Affects the maximum number of concurrent queries. Usually, the API server
                     status includes a number of slots it provides for each IP. If that is the
                     case, we pick the minimum of that number of slots and ``concurrency`` as
                     concurrency limit. If the server does not provide a limit itself,
                     ``concurrency`` will be used as concurrency limit.
        runner: You can provide another query runner if you want to implement your own retry
                strategy.

    References:
        - https://wiki.openstreetmap.org/wiki/Overpass_API#Public_Overpass_API_instances
    """

    __slots__ = (
        "_concurrency",
        "_maybe_any_status",
        "_maybe_sem",
        "_maybe_session",
        "_runner",
        "_url",
        "_user_agent",
    )

    def __init__(
        self,
        url: str = DEFAULT_INSTANCE,
        user_agent: str = DEFAULT_USER_AGENT,
        concurrency: int = 32,
        runner: QueryRunner | None = None,
    ) -> None:
        if concurrency <= 0:
            msg = "'concurrency' must be > 0"
            raise ValueError(msg)

        self._url = url
        self._user_agent = user_agent
        self._concurrency = concurrency
        self._runner = runner or DefaultQueryRunner()

        self._maybe_session: aiohttp.ClientSession | None = None
        self._maybe_any_status: Status | None = None
        self._maybe_sem: asyncio.BoundedSemaphore | None = None

    def _session(self) -> aiohttp.ClientSession:
        """The session used for all requests of this client."""
        if not self._maybe_session or self._maybe_session.closed:
            headers = {"User-Agent": self._user_agent}
            connector = aiohttp.TCPConnector(limit=0)  # we limit concurrency ourselves
            self._maybe_session = aiohttp.ClientSession(headers=headers, connector=connector)

        return self._maybe_session

    async def _rate_limiter(
        self,
        timeout: ClientTimeout | object = sentinel,
    ) -> asyncio.BoundedSemaphore:
        """
        A rate-limiting semaphore for queries.

        This semaphore can be acquired as many times as there are query slots at the API instance.
        If all slots are acquired, further queries need to wait until a slot is released.
        """
        status = self._maybe_any_status

        if self._maybe_sem:
            return self._maybe_sem

        if not status:
            status = await self._status(timeout=timeout)

        self._maybe_sem = asyncio.BoundedSemaphore(status.concurrency)
        return self._maybe_sem

    async def close(self) -> None:
        """Cancel all running queries and close the underlying session."""
        if self._maybe_session and not self._maybe_session.closed:
            # do not care if this fails
            with suppress(CallError):
                _ = await self.cancel_queries()

            # is raised when there are still active queries. that's ok
            with suppress(aiohttp.ServerDisconnectedError):
                await self._maybe_session.close()

    async def _status(self, timeout: ClientTimeout | object = sentinel) -> "Status":
        try:
            async with self._session().get(
                url=urljoin(self._url, "status"), timeout=timeout
            ) as response:
                return await self._status_from_response(response)
        except aiohttp.ClientError as err:
            raise await _to_client_error(err) from err

    async def _status_from_response(self, response: aiohttp.ClientResponse) -> "Status":
        text = await response.text()

        slots: int | None = 0
        free_slots = None
        cooldown_secs = 0
        concurrency = self._concurrency
        endpoint = None
        nb_running_queries = 0

        try:
            match_slots_overall = re.findall(r"Rate limit: (\d+)", text)
            match_slots_available = re.findall(r"(\d+) slots available now", text)
            match_cooldowns = re.findall(r"Slot available after: .+, in (\d+) seconds", text)
            match_endpoint = re.findall(r"Announced endpoint: (.+)", text)
            match_running_queries = re.findall(
                r"\d+\t\d+\t\d+\t\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", text
            )

            (slots_str,) = match_slots_overall
            slots = int(slots_str) or None

            endpoint = match_endpoint[0] if match_endpoint else None
            endpoint = None if endpoint == "none" else endpoint

            nb_running_queries = len(match_running_queries)

            if slots:
                cooldowns = [int(secs) for secs in match_cooldowns]

                if match_slots_available:
                    free_slots = int(match_slots_available[0])
                else:
                    free_slots = slots - len(cooldowns)

                cooldown_secs = 0 if free_slots > 0 else min(cooldowns)

                # pick the server's concurrent query limit if > 0 and < self._concurrency,
                # or self._concurrency otherwise
                concurrency = min(slots or self._concurrency, self._concurrency)
        except ValueError as err:
            raise await _to_client_error(response) from err

        self._maybe_any_status = Status(
            slots=slots,
            free_slots=free_slots,
            cooldown_secs=cooldown_secs,
            concurrency=concurrency,
            endpoint=endpoint,
            nb_running_queries=nb_running_queries,
        )
        return self._maybe_any_status

    async def status(self) -> Status:
        """
        Check the current API status.

        Raises:
            ClientError: if the status could not be looked up
        """
        return await self._status()

    async def cancel_queries(self) -> int:
        """
        Cancel all running queries.

        This can be used to terminate runaway queries that prevent you from sending new ones.

        Returns:
            the number of terminated queries

        Raises:
            ClientError: if the request to cancel queries failed
        """
        session = self._session()
        endpoint = urljoin(self._url, "kill_my_queries")
        try:
            async with session.get(endpoint) as response:
                body = await response.text()
                killed_pids = re.findall("\\(pid (\\d+)\\)", body)
                return len(set(killed_pids))
        except aiohttp.ClientError as err:
            raise await _to_client_error(err) from err

    async def run_query(self, query: Query, raise_on_failure: bool = True) -> None:
        """
        Send a query to the API, and await its completion.

        "Running" the query entails acquiring a connection from the pool, waiting for a slot
        to open up, the query requests themselves (which may be retried), status requests
        when the server is busy, and cooldown periods.

        The query runner is invoked before every try, and once after the last try.

        To run multiple queries concurrently, wrap the returned coroutines in an ``asyncio`` task,
        f.e. with ``asyncio.create_task()`` and subsequent ``asyncio.gather()``.

        Args:
            query: the query to run on this API instance
            raise_on_failure: if ``True``, raises ``query.error`` if the query failed

        Raises:
            ClientError: when query or status requests fail. If the query was retried, the error
                         of the last try will be raised. The same exception is also captured in
                         ``query.error``. Raising can be prevented by setting ``raise_on_failure``
                         to ``False``.
            RunnerError: when a call to the query runner raises. This exception is raised
                         even if ``raise_on_failure` is ``False``, since it is likely an error
                         that is not just specific to this query.
        """
        if query.done:
            return  # nothing to do

        if query.nb_tries > 0:
            query.reset()  # reset failed queries

        while True:
            await self._invoke_runner(query, raise_on_failure=raise_on_failure)
            if query.done:
                return
            await self._run_query_once(query)

    async def _invoke_runner(self, query: Query, raise_on_failure: bool) -> None:
        try:
            await self._runner(query)
        except ClientError as err:
            if err is not query.error:
                msg = "query runner raised a ClientError other than 'query.error'"
                raise ValueError(msg) from err
            if raise_on_failure:
                raise
        except BaseException as err:
            raise RunnerError(err) from err

    async def _run_query_once(self, query: Query) -> None:
        logger = query.logger

        if query.done:
            return

        query_mut = query._mutator()

        query_mut.begin_try()

        acquired_slot = False

        if query.request_timeout.total_without_query_secs is not None:
            total = float(query.timeout_secs) + query.request_timeout.total_without_query_secs
        else:
            total = None

        req_timeout = aiohttp.ClientTimeout(
            total=total,
            connect=None,
            sock_connect=query.request_timeout.sock_connect_secs,
            sock_read=query.request_timeout.each_sock_read_secs,
        )

        try:
            await self._wait_for_slot(query)
            acquired_slot = True

            query_mut.begin_request()

            logger.info(f"call api for {query}")

            async with self._session().get(
                url=urljoin(self._url, "interpreter"),
                params={"data": query.code},
                timeout=req_timeout,
            ) as response:
                query_mut.succeed_try(
                    response=await _result_or_raise(response, query.kwargs),
                    response_bytes=response.content.total_bytes,
                )

        except aiohttp.ClientError as err:
            query_mut.fail_try(await _to_client_error(err))

        except asyncio.TimeoutError as err:
            query_err: ClientError

            if query.run_timeout_elapsed:
                assert query.run_duration_secs is not None
                query_err = GiveupError(kwargs=query.kwargs, after_secs=query.run_duration_secs)
            else:
                assert not query.run_timeout_secs or acquired_slot
                assert req_timeout.total
                query_err = CallTimeoutError(cause=err, after_secs=req_timeout.total)

            query_mut.fail_try(query_err)

        except ClientError as err:
            query_mut.fail_try(err)

        finally:
            query_mut.end_try()
            if acquired_slot:
                assert self._maybe_sem is not None
                self._maybe_sem.release()

    async def _wait_for_slot(self, query: Query) -> None:
        def next_timeout() -> aiohttp.ClientTimeout:
            assert query.run_duration_secs is not None
            if query.run_timeout_secs:
                remaining = query.run_timeout_secs - query.run_duration_secs
                if remaining <= 0.0:
                    raise asyncio.TimeoutError()  # no point delaying the inevitable
            else:
                remaining = None  # no limit

            return aiohttp.ClientTimeout(total=remaining)

        logger = query.logger

        check_cooldown = (
            isinstance(query.error, QueryRejectError)
            and query.error.cause == QueryRejectCause.TOO_MANY_QUERIES
        )

        if check_cooldown:
            # If this client is running too many queries, we can check the status for a
            # cooldown period. This request failing is a bit of an edge case.
            # 'query.error' will be overwritten, which means we will not check for a
            # cooldown in the next iteration.
            status = await self._status(timeout=next_timeout())

            if (
                (timeout := next_timeout())
                and timeout.total is not None
                and status.cooldown_secs > timeout.total
            ):
                assert query.run_duration_secs
                raise GiveupError(kwargs=query.kwargs, after_secs=query.run_duration_secs)

            logger.info(f"{query} has cooldown for {status.cooldown_secs:.1f}s")
            await asyncio.sleep(status.cooldown_secs)

        # This requests an API status if we haven't done so already.
        rate_limiter = await self._rate_limiter(timeout=next_timeout())

        # Limit the concurrent query requests to the number of slots available.
        if rate_limiter.locked():
            logger.info(f"{query} has to wait for a slot")

        await asyncio.wait_for(
            fut=rate_limiter.acquire(),
            timeout=next_timeout().total,
        )
